<?php

  require_once 'Tree.class.php';

  class TreeParser
  {

    // special symbols used in Newick format
    static $symbols = array( 'CHILD_LIST_START'     => '(',
                             'CHILD_LIST_END'       => ')',
                             'CHILD_LIST_SEPARATOR' => ',',
                             'COMPLETE_TREE_END'    => ';',
                             'BRANCH_LENGTH_SIGN'   => ':'  );

    private $text   = null;  // input: text representation of the trees to be parsed
    private $tokens = null;  // list of all tokens in the text
    private $trees  = null;  // output: array of the parsed trees
    private $error  = null;  // parse error message
    private $curtok = null;  // current token number (cursor)

    // constructor
    public function TreeParser( $text=null )
    {
      $this->setInput( $text );
    }

    // resets the object and stores a new tree text to parse
    public function setInput( $text=null )
    {
      $this->reset();
      $this->text = $text;
    }

    // resets the object to it's initial state (cleans everything up)
    public function reset()
    {
      $this->text   = null;
      $this->tokens = null;
      $this->trees  = null;
      $this->error  = null;
      $this->curtok = null;
    }

    // returns an array of the parsed trees
    public function getTrees()
    {

      if ( $this->trees === null )
      {

        if ( $this->tokens === null )
          $this->tokenize();

        $this->curtok = 0;
        $this->trees = $this->parseTrees();

        $this->curtok = null;
        $this->tokens = null; // frees much memory

      }

      return $this->trees;

    }

    public function errorOccured()
    {
      return !empty($this->error);
    }

    public function getErrorMessage()
    {
      return $this->error;
    }

    // {tree_list}     : {complete_tree}*
    // {complete_tree} : {tree};
    // {tree}          : {children}{label}
    // {children}      ; (\({children_list}\))?
    // {children_list} : {tree}(,{tree})*
    // {label}         : T_LABEL?

    // tries to parse a list of trees,
    // returns an array of successfully parsed trees
    private function parseTrees()
    {
      $trees = array();
      while ( !$this->error and !$this->outOfTokens() )
      {
         $tree = $this->parseCompleteTree();
         if ( !$this->error and $tree !== false ) // a tree has been read successfully
           $trees[] = $tree;
      }
      return $trees;
    }

    // tries to parse a complete tree (closed by ';'),
    // returns the tree or false in case of an error
    private function parseCompleteTree()
    {

      if ( !$this->check4MoreTokens() )
        return false;

      $info = $this->parseLabel();
      if ( $this->error or $this->outOfTokens() )
        return false;

      $tree = $this->parseTree();
      if ( $this->error )
        return false;

      if ( !$this->expectTokenType( TreeToken::TREE_END ) )
        return false;

      $this->go2NextToken();

      // this makes 'A' in Newick tree "A;" a label rather than an info tag
      if ( !empty($info) )
        if ( $tree->hasChildren() )
          $tree->setInfo($info);
        else
          $tree->setLabel($info);

      return $tree;

    }

    // tries to parse a (sub)tree,
    // returns a tree or false in case of an error
    private function parseTree()
    {

      if ( !$this->check4MoreTokens() )
        return false;

      $tree = new Tree();

      $children = $this->parseChildren();
      if ( $this->error )
        return false;

      $label = $this->parseLabel();
      if ( $this->error )
        return false;

      if ( count($children) )
        $tree->addChildren($children);
      if ( !empty($label) )
        $tree->setLabel( $label );

      return $tree;

    }

    // tries to parse a (possibly empty) list of subtrees,
    // returns an array of subtrees or false in case of an error
    private function parseChildren()
    {

      if ( !$this->check4MoreTokens() )
        return false;

      $children = array();

      if ( $this->currentTokenType() !== TreeToken::LIST_START )
        return $children;

      $this->go2NextToken();

      $children = $this->parseChildrenList();
      if ( $this->error )
        return false;

      if ( !$this->check4MoreTokens() )
        return false;

      if ( !$this->expectTokenType( TreeToken::LIST_END ) )
        return false;

      $this->go2NextToken();

      return $children;

    }

    // tries to parse a non-empty list of subtrees,
    // returns an array of subtrees or false in case of an error
    private function parseChildrenList()
    {

      if ( !$this->check4MoreTokens() )
        return false;

      $child = $this->parseTree();
      if ( $this->error )
        return false;

      if ( !$this->check4MoreTokens() )
        return false;

      if ( $this->currentTokenType() !== TreeToken::LIST_SEPARATOR )
        return array( $child );

      $this->go2NextToken();

      $children = $this->parseChildrenList();
      if ( $this->error )
        return false;

      array_unshift( $children, $child );

      return $children;

    }

    // parses a lable (may be empty),
    // returns a string or false in case of an error
    private function parseLabel()
    {
      $label = "";
      if ( $this->currentTokenType() === TreeToken::LABEL )
      {
        $label = $this->currentTokenText();
        $this->go2NextToken();
      }
      return $label;
    }

    // prepares a list of tokens
    private function tokenize()
    {
      $this->tokens = array();
      $tl = strlen($this->text);
      for ( $i=0; $i<$tl; $i++ )
        switch ( $char = $this->text[$i] )
        {
        case self::$symbols['CHILD_LIST_START']:
          $this->tokens[] = new TreeToken( TreeToken::LIST_START, $char );
          break;
        case self::$symbols['CHILD_LIST_END']:
          $this->tokens[] = new TreeToken( TreeToken::LIST_END, $char );
          break;
        case self::$symbols['CHILD_LIST_SEPARATOR']:
          $this->tokens[] = new TreeToken( TreeToken::LIST_SEPARATOR, $char );
          break;
        case self::$symbols['COMPLETE_TREE_END']:
          $this->tokens[] = new TreeToken( TreeToken::TREE_END, $char );
          break;
        case ' ': case "\t": case "\n": case "\r": break; // ignore whitespace
        default:
          $label = '';
          $stop_label = false;
          do
            switch ( $char = $this->text[$i] )
            {
              case self::$symbols['CHILD_LIST_START']:
              case self::$symbols['CHILD_LIST_END']:
              case self::$symbols['CHILD_LIST_SEPARATOR']:
              case self::$symbols['COMPLETE_TREE_END']:
                $stop_label = true;
                --$i; // step back 'coz it's special sign
                break;
              case ' ': case "\t": case "\n": case "\r":
                if ( $label[strlen($label)-1] !== ' ' )
                  $label .= ' ';
                break;
              default:
                $label .= $char;
                break;
            }
          while ( !$stop_label and ++$i<$tl );
          $label = trim($label);
          if ( !empty($label) )
            $this->tokens[] = new TreeToken( TreeToken::LABEL, $label );
          break;
        }
    }

    private function outOfTokens()
    {
      return ( $this->curtok >= count($this->tokens) );
    }

    private function go2NextToken()
    {
      if ( !$this->outOfTokens() )
        ++ $this->curtok;
    }

    private function check4MoreTokens()
    {
      if ( $this->outOfTokens() )
        $this->error = "Unexpected end of input.";
      return !$this->errorOccured();
    }

    private function currentToken()
    {
      return ( $this->outOfTokens() ?
               false :
               $this->tokens[$this->curtok] );
    }

    private function currentTokenType()
    {
      $token = $this->currentToken();
      return ( $token === false ? false : $token->type );
    }

    private function currentTokenText()
    {
      $token = $this->currentToken();
      return ( $token === false ? false : $token->text );
    }

    private function expectTokenType( $type )
    {
      $found = $this->currentTokenType();
      if ( $found !== $type )
        $this->error = "Syntax error,".
                       " unexpected " . TreeToken::getTypeName($found) .
                       ", expecting " . TreeToken::getTypeName($type) .
                       " near \"" . $this->currentNeighbourhood() . "\".";
      return !$this->errorOccured();
    }

    private function currentNeighbourhood( $from=0, $to=9 )
    {
      $neigh = "";
      for ( $i=$from; $i<=$to; $i++ )
        if ( $this->curtok+$i > 0 and $this->curtok+$i < count($this->tokens) )
          $neigh .= $this->tokens[$this->curtok+$i]->text;
      return $neigh;
    }

  } // class TreeParser

  class TreeToken
  {

    const LABEL          = '*';
    const LIST_START     = '(';
    const LIST_END       = ')';
    const LIST_SEPARATOR = ',';
    const TREE_END       = ';';
    static $token_names = array(
      '*' => 'LABEL',
      '(' => 'LIST_START',
      ')' => 'LIST_END',
      ',' => 'LIST_SEPARATOR',
      ';' => 'TREE_END' );

    var $type = null; //  
    var $text = null; //  

    function TreeToken( $type=null, $text=null )
    {
      $this->type = $type;
      $this->text = $text;
    }

    static function getTypeName( $type )
    {
      return self::$token_names[$type];
    }

  } // class TreeToken

?>
